import logging
import pathlib
import warnings
from typing import Any, Literal, Optional, Union

import pystac_client
import requests

from up42 import host, utils
from up42.http import client, oauth

logger = utils.get_logger(__name__, level=logging.INFO)

warnings.simplefilter(action="ignore", category=FutureWarning)


class UserNotAuthenticated(ValueError):
    pass


def _authenticated(value: Any):
    if value:
        return value
    raise UserNotAuthenticated("User not authenticated.")


class _Workspace:
    _id: Optional[str] = None
    _session: Optional[requests.Session] = None
    _auth: Optional[oauth.Up42Auth] = None

    @property
    def id(self):
        return _authenticated(self._id)

    @property
    def session(self):
        return _authenticated(self._session)

    @property
    def auth(self):
        return _authenticated(self._auth)

    def authenticate(
        self,
        cfg_file: Optional[Union[str, pathlib.Path]] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        region: Literal["eu", "sa"] = "eu",
    ):
        """
        Authenticate with UP42, either using account credentials or a config JSON file
        containing the corresponding credentials.

        Args:
            cfg_file: File path to the cfg.json with {username: "...", password: "..."}.
            username: The username for the UP42 account (email UP42 console).
            password: Password for the UP42 console login.
            region: The desired region to use for all SDK operations.
        """
        host.REGION = region
        credential_sources = client.collect_credentials(cfg_file, username, password)
        up42_client = client.create(credential_sources, host.token_endpoint())
        logger.info("Authentication with UP42 successful!")
        self._session = up42_client.session
        user_info = self.session.get(host.user_info_endpoint()).json()
        self._id = user_info["sub"]
        self._auth = up42_client.auth

    def get_credits_balance(self) -> dict:
        """
        Display the overall credits available in your account.

        Returns:
            A dict with the balance of credits available in your account.
        """
        url = host.endpoint("/v2/payments/balances")
        balance = self.session.get(url).json()["available"]["amount"]
        return {"balance": balance}


workspace = _Workspace()

authenticate = workspace.authenticate
get_credits_balance = workspace.get_credits_balance


def stac_client():
    return utils.stac_client(workspace.auth)


class Session:
    def __get__(self, obj, obj_type=None) -> requests.Session:
        return workspace.session


class WorkspaceId:
    def __get__(self, obj, obj_type=None) -> str:
        if obj:
            return obj.__dict__.get("workspace_id", workspace.id)
        return workspace.id

    def __set__(self, obj, value: str) -> None:
        if value == self:
            value = workspace.id
        obj.__dict__["workspace_id"] = value


class StacClient:
    def __get__(self, obj, obj_type=None) -> pystac_client.Client:
        return stac_client()



"""
SDK conveniance functionality that is made available on the up42 import object (in the init) and is not directly
related to API calls.
"""
import contextlib
import logging
import pathlib
from typing import Dict, Union

import geopandas  # type: ignore
import pandas
import shapely  # type: ignore

from up42 import utils

logger = utils.get_logger(__name__)


def read_vector_file(filename: str = "aoi.geojson", as_dataframe: bool = False) -> Union[Dict, geopandas.GeoDataFrame]:
    """
    Reads vector files (geojson, shapefile, kml, wkt) to a feature collection,
    for use as the AOI geometry in the workflow input parameters
    (see get_input_parameters).

    Example AOI fields are provided, e.g. example/data/aoi_Berlin.geojson

    Args:
        filename: File path of the vector file.
        as_dataframe: Return type, default FeatureCollection, GeoDataFrame if True.

    Returns:
        Feature Collection
    """
    suffix = pathlib.Path(filename).suffix

    if suffix == ".kml":
        # pylint: disable=no-member
        with contextlib.suppress(AttributeError):
            # This is needed for geopandas < 1
            geopandas.io.file.fiona.drvsupport.supported_drivers["KML"] = "rw"
        df = geopandas.read_file(filename, driver="KML")
    elif suffix == ".wkt":
        with open(filename, encoding="utf-8") as wkt_file:
            wkt = wkt_file.read()
            df = pandas.DataFrame({"geometry": [wkt]})
            df["geometry"] = df["geometry"].apply(shapely.wkt.loads)
            df = geopandas.GeoDataFrame(df, geometry="geometry", crs=4326)
    else:
        df = geopandas.read_file(filename)

    if df.crs.to_string() != "EPSG:4326":
        df = df.to_crs(epsg=4326)
    if as_dataframe:
        return df
    else:
        return df.__geo_interface__


def get_example_aoi(location: str = "Berlin", as_dataframe: bool = False) -> Union[Dict, geopandas.GeoDataFrame]:
    """
    Gets predefined, small, rectangular example AOI for the selected location.

    Args:
        location: Location, one of Berlin, Washington.
        as_dataframe: Returns a dataframe instead of dict FeatureColletions
            (default).

    Returns:
        Feature collection JSON with the selected AOI.
    """
    logger.info("Getting small example AOI in location '%s'.", location)
    if location == "Berlin":
        example_aoi = read_vector_file(f"{str(pathlib.Path(__file__).resolve().parent)}/data/aoi_berlin.geojson")
    elif location == "Washington":
        example_aoi = read_vector_file(f"{str(pathlib.Path(__file__).resolve().parent)}/data/aoi_washington.geojson")
    else:
        raise ValueError("Please select one of 'Berlin' or 'Washington' as the location!")

    if as_dataframe:
        df = geopandas.GeoDataFrame.from_features(example_aoi, crs=4326)
        return df
    else:
        return example_aoi


def settings(log: bool = True) -> None:
    """
    Configures settings about logging etc. when using the up42-py package.
    Args:
        log: Activates/deactivates logging, default True is activated logging.
    """
    if log:
        logger.info("Logging enabled (default) - use up42.settings(log=False) to disable.")
    else:
        logger.info("Logging disabled - use up42.settings(log=True) to reactivate.")

    for name in logging.root.manager.loggerDict:
        setattr(logging.getLogger(name), "disabled", not log)



"""
    `up42` is the base library module imported to Python. It provides the elementary
    functionality that is not bound to a specific class of the UP42 structure.
    From `up42` you can also initialize other classes, e.g. for using the catalog, storage etc.

    To import the UP42 library:
    ```python
    import up42
    ```

    Authenticate with UP42:
        https://sdk.up42.com/authentication/.

    To initialize any lower level functionality use e.g.
    ```python
    catalog = up42.initialize_catalog()
    ```
"""

from typing import Callable, Type, Union, cast

# pylint: disable=only-importing-modules-is-allowed
from up42.asset import Asset, AssetSorting
from up42.base import authenticate, get_credits_balance, stac_client
from up42.catalog import Catalog
from up42.glossary import CollectionSorting, CollectionType, ProductGlossary
from up42.initialization import (
    initialize_asset,
    initialize_catalog,
    initialize_order,
    initialize_storage,
    initialize_tasking,
)
from up42.order import Order, OrderSorting
from up42.order_template import BatchOrderTemplate
from up42.processing import Job, JobSorting, JobStatus
from up42.stac import extend as stac_extend
from up42.storage import Storage
from up42.tasking import Quotation, QuotationSorting, Tasking
from up42.tools import get_example_aoi, read_vector_file
from up42.utils import get_up42_py_version
from up42.webhooks import Webhook

stac_extend()

__version__ = get_up42_py_version()
__all__ = [
    cast(
        Union[Type, Callable],
        obj,
    ).__name__
    for obj in [
        Asset,
        AssetSorting,
        Catalog,
        Order,
        OrderSorting,
        Storage,
        Tasking,
        Webhook,
        get_example_aoi,
        read_vector_file,
        initialize_catalog,
        initialize_tasking,
        initialize_storage,
        initialize_order,
        initialize_asset,
        authenticate,
        get_credits_balance,
        stac_client,
        Job,
        JobSorting,
        JobStatus,
        CollectionSorting,
        CollectionType,
        ProductGlossary,
        BatchOrderTemplate,
        Quotation,
        QuotationSorting,
    ]
]



import copy
import dataclasses
from typing import Any, Dict, Iterator, List, Literal, Optional, TypedDict, Union, cast

import tenacity as tnc

from up42 import asset, base, host, utils

logger = utils.get_logger(__name__)

MAX_ITEM = 200
LIMIT = 200

OrderType = Literal["TASKING", "ARCHIVE"]


class OrderParams(TypedDict, total=False):
    """
    Represents the schema for the order parameters.
    dataProduct: The dataProduct id for the specific product configuration.
    params: Order parameters for each product. \
        They are different from product to product depending on product schema.
    tags: User tags to helping to identify the order.
    """

    dataProduct: str  # pylint: disable=invalid-name
    params: Dict[str, Any]
    tags: List[str]


class OrderParamsV2(TypedDict, total=False):
    """
    Represents the schema for the order parameters for the V2 endpoint.
    dataProduct: The dataProduct id for the specific product configuration.
    displayName: The default name that will be identifying the order.
    featureCollection: The AOI of the order.
    params: Order parameters for each product.
    tags: User tags to helping to identify the order.
    """

    # pylint: disable=invalid-name
    dataProduct: str
    displayName: str
    params: Dict[str, Any]
    featureCollection: Dict[str, Any]
    tags: List[str]


def _translate_construct_parameters(order_parameters: OrderParams) -> OrderParamsV2:
    order_parameters_v2 = cast(OrderParamsV2, copy.deepcopy(order_parameters))
    params = order_parameters_v2["params"]
    data_product_id = order_parameters_v2["dataProduct"]
    order_parameters_v2["displayName"] = params.get("displayName", f"{data_product_id} order")
    aoi = params.pop("aoi", None) or params.pop("geometry", None)
    order_parameters_v2["featureCollection"] = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "geometry": aoi,
            }
        ],
    }
    return order_parameters_v2


OrderStatus = Literal[
    "CREATED",
    "BEING_PLACED",
    "PLACED",
    "BEING_FULFILLED",
    "DELIVERY_INITIALIZATION_FAILED",
    "DOWNLOAD_FAILED",
    "DOWNLOADED",
    "FULFILLED",
    "FAILED",
    "FAILED_PERMANENTLY",
]
OrderSubStatus = Literal[
    "FEASIBILITY_WAITING_UPLOAD",
    "FEASIBILITY_WAITING_RESPONSE",
    "QUOTATION_WAITING_UPLOAD",
    "QUOTATION_WAITING_RESPONSE",
    "QUOTATION_ACCEPTED",
    "QUOTATION_REJECTED",
]


class UnfulfilledOrder(ValueError):
    pass


class FailedOrder(ValueError):
    pass


class FailedOrderPlacement(ValueError):
    pass


class OrderSorting:
    created_at = utils.SortingField(name="createdAt")
    updated_at = utils.SortingField(name="updatedAt")
    type = utils.SortingField(name="type")
    status = utils.SortingField(name="status")


@dataclasses.dataclass
class ArchiveOrderDetails:
    aoi: dict
    image_id: Optional[str]
    sub_status = None


@dataclasses.dataclass
class TaskingOrderDetails:
    acquisition_start: str
    acquisition_end: str
    geometry: dict
    extra_description: Optional[str]
    sub_status: Optional[OrderSubStatus]


OrderDetails = Union[ArchiveOrderDetails, TaskingOrderDetails]


@dataclasses.dataclass
class Order:
    session = base.Session()
    id: str
    display_name: str
    status: OrderStatus
    workspace_id: str
    account_id: str
    type: OrderType
    details: Optional[OrderDetails]
    data_product_id: Optional[str]
    tags: Optional[list[str]]
    info: dict = dataclasses.field(repr=False)

    @classmethod
    def get(cls, order_id: str) -> "Order":
        url = host.endpoint(f"/v2/orders/{order_id}")
        metadata = cls.session.get(url=url).json()
        return Order._from_metadata(metadata)

    @staticmethod
    def _from_metadata(data: dict) -> "Order":
        details: Optional[OrderDetails] = None
        if "orderDetails" in data:
            order_details: dict = data["orderDetails"]
            if data["type"] == "TASKING":
                details = TaskingOrderDetails(
                    acquisition_start=order_details["acquisitionStart"],
                    acquisition_end=order_details["acquisitionEnd"],
                    geometry=order_details["geometry"],
                    extra_description=order_details.get("extraDescription"),
                    sub_status=order_details.get("subStatus"),
                )
            else:
                details = ArchiveOrderDetails(aoi=order_details["aoi"], image_id=order_details.get("imageId"))
        return Order(
            id=data["id"],
            display_name=data["displayName"],
            status=data["status"],
            workspace_id=data["workspaceId"],
            account_id=data["accountId"],
            type=data["type"],
            details=details,
            data_product_id=data.get("dataProductId"),
            tags=data.get("tags"),
            info=data,
        )

    @classmethod
    def all(
        cls,
        workspace_id: Optional[str] = None,
        order_type: Optional[OrderType] = None,
        status: Optional[List[OrderStatus]] = None,
        sub_status: Optional[List[OrderSubStatus]] = None,
        display_name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        sort_by: Optional[utils.SortingField] = None,
    ) -> Iterator["Order"]:
        params = {
            "sort": sort_by,
            "workspaceId": workspace_id,
            "displayName": display_name,
            "type": order_type,
            "tags": tags,
            "status": status,
            "subStatus": sub_status,
        }
        return map(cls._from_metadata, utils.paged_query(params, "/v2/orders", cls.session))

    @property
    @utils.deprecation("Order.details", "3.0.0")
    def order_details(self) -> dict:
        return self.info.get("orderDetails", {})

    @property
    @utils.deprecation("Order.id", "3.0.0")
    def order_id(self) -> str:
        return self.info["id"]

    @property
    def is_fulfilled(self) -> bool:
        """
        Gets `True` if the order is fulfilled, `False` otherwise.
        Also see [status attribute](order-reference.md#up42.order.Order.status).
        """
        return self.status == "FULFILLED"

    def get_assets(self) -> List[asset.Asset]:
        """
        Gets the Order assets or results.
        """
        if self.status not in ["FULFILLED", "BEING_FULFILLED"]:
            raise UnfulfilledOrder(f"""Order {self.order_id} is not valid. Current status is {self.status}""")
        return list(asset.Asset.all(search=self.order_id))

    @classmethod
    @utils.deprecation("OrderTemplate::place", "3.0.0")
    def place(cls, order_parameters: OrderParams, workspace_id: str) -> "Order":
        url = host.endpoint(f"/v2/orders?workspaceId={workspace_id}")
        response_json = cls.session.post(url=url, json=_translate_construct_parameters(order_parameters)).json()
        if response_json["errors"]:
            message = response_json["errors"][0]["message"]
            raise FailedOrderPlacement(f"Order was not placed: {message}")
        order = cls.get(response_json["results"][0]["id"])
        logger.info("Order %s is now %s.", order.order_id, order.status)
        return order

    @classmethod
    @utils.deprecation("OrderTemplate.estimate", "3.0.0")
    def estimate(cls, order_parameters: OrderParams) -> int:
        """
        Returns an estimation of the cost of an order.

        Args:
            auth: An authentication object.
            order_parameters: A dictionary for the order configuration.

        Returns:
            int: The estimated cost of the order
        """

        url = host.endpoint("/v2/orders/estimate")
        response_json = cls.session.post(
            url=url,
            json=_translate_construct_parameters(order_parameters),
        ).json()
        estimated_credits: int = response_json["summary"]["totalCredits"]
        logger.info(
            "Order is estimated to cost %s UP42 credits (order_parameters: %s)",
            estimated_credits,
            order_parameters,
        )
        return estimated_credits

    def track(self, report_time: float = 120):
        logger.info("Tracking order updates, reporting every %s seconds...", report_time)

        @tnc.retry(
            wait=tnc.wait_fixed(report_time),
            retry=tnc.retry_if_exception_type(UnfulfilledOrder),
            reraise=True,
        )
        def update():
            order = Order.get(self.id)
            for field in dataclasses.fields(order):
                setattr(self, field.name, getattr(order, field.name))
            sub_status = self.details and self.details.sub_status
            sub_status_msg = f": {sub_status}" if sub_status is not None else ""

            logger.info("Order is %s! - %s", self.status + sub_status_msg, self.order_id)
            if self.status in ["FAILED", "FAILED_PERMANENTLY"]:
                raise FailedOrder("Order has failed!")
            if not self.is_fulfilled:
                raise UnfulfilledOrder

        update()

    @utils.deprecation("Order::track", "3.0.0")
    def track_status(self, report_time: float = 120) -> str:
        self.track(report_time)
        return self.status



import dataclasses
from typing import List, Optional, Union

import pystac

from up42 import base, processing


# TODO: drop these with Python 3.10 kw_only=True data classes
@dataclasses.dataclass
class WorkspaceIdSingleItemTemplate(processing.SingleItemJobTemplate):
    workspace_id: Union[str, base.WorkspaceId] = dataclasses.field(default=base.WorkspaceId())


@dataclasses.dataclass
class WorkspaceIdMultiItemTemplate(processing.MultiItemJobTemplate):
    workspace_id: Union[str, base.WorkspaceId] = dataclasses.field(default=base.WorkspaceId())


@dataclasses.dataclass
class DetectionBuildingsSpacept(WorkspaceIdSingleItemTemplate):
    process_id = "detection-buildings-spacept"


@dataclasses.dataclass
class DetectionTreesSpacept(WorkspaceIdSingleItemTemplate):
    process_id = "detection-trees-spacept"


@dataclasses.dataclass
class DetectionShadowsSpacept(WorkspaceIdSingleItemTemplate):
    process_id = "detection-shadows-spacept"


@dataclasses.dataclass
class DetectionShipsAirbus(WorkspaceIdSingleItemTemplate):
    process_id = "detection-ships-airbus"


@dataclasses.dataclass
class DetectionWindTurbinesAirbus(WorkspaceIdSingleItemTemplate):
    process_id = "detection-wind-turbines-airbus"


@dataclasses.dataclass
class DetectionStorageTanksAirbus(WorkspaceIdSingleItemTemplate):
    process_id = "detection-storage-tanks-airbus"


@dataclasses.dataclass
class DetectionTrucksOI(WorkspaceIdSingleItemTemplate):
    process_id = "detection-trucks-oi"


@dataclasses.dataclass
class DetectionCarsOI(WorkspaceIdSingleItemTemplate):
    process_id = "detection-cars-oi"


@dataclasses.dataclass
class DetectionAircraftOI(WorkspaceIdSingleItemTemplate):
    process_id = "detection-aircraft-oi"


@dataclasses.dataclass
class DetectionChangeSpacept(WorkspaceIdMultiItemTemplate):
    process_id = "detection-change-spacept"


@dataclasses.dataclass
class DetectionChangePleiadesHyperverge(WorkspaceIdMultiItemTemplate):
    process_id = "detection-change-pleiades-hyperverge"


@dataclasses.dataclass
class DetectionChangeSPOTHyperverge(WorkspaceIdMultiItemTemplate):
    process_id = "detection-change-spot-hyperverge"


@dataclasses.dataclass
class CoregistationSimularity(processing.JobTemplate):
    title: str
    source_item: pystac.Item
    reference_item: pystac.Item
    process_id = "coregistration-simularity"
    workspace_id: Union[str, base.WorkspaceId] = dataclasses.field(default=base.WorkspaceId())

    @property
    def inputs(self) -> dict:
        return {
            "title": self.title,
            "sourceItem": self.source_item.get_self_href(),
            "referenceItem": self.reference_item.get_self_href(),
        }


@dataclasses.dataclass
class UpsamplingNS(processing.SingleItemJobTemplate):
    ned: bool = False
    rgb: bool = True
    workspace_id: Union[str, base.WorkspaceId] = dataclasses.field(default=base.WorkspaceId())
    process_id = "upsampling-ns"

    @property
    def inputs(self):
        return {
            **super().inputs,
            "NED": self.ned,
            "RGB": self.rgb,
        }


@dataclasses.dataclass
class GreyWeight:
    band: str
    weight: float


@dataclasses.dataclass
class Pansharpening(processing.SingleItemJobTemplate):
    grey_weights: Optional[List[GreyWeight]] = None
    workspace_id: Union[str, base.WorkspaceId] = dataclasses.field(default=base.WorkspaceId())
    process_id = "pansharpening"

    @property
    def inputs(self):
        weights = (
            {"greyWeights": [dataclasses.asdict(weight) for weight in self.grey_weights]} if self.grey_weights else {}
        )
        return {**super().inputs, **weights}



import dataclasses
from typing import List, Optional

from up42 import base, host, utils

logger = utils.get_logger(__name__)


@dataclasses.dataclass
class Webhook:
    session = base.Session()
    workspace_id = base.WorkspaceId()
    url: str
    name: str
    events: List[str]
    active: bool = False
    secret: Optional[str] = None
    id: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None

    """
    # Webhook
    Contains UP42 webhooks functionality to set up a custom callback e.g. when an order is finished
    a webhook is triggered and an event notification is transmitted via HTTPS to a specific URL.

    Also see the [full webhook documentation](https://docs.up42.com/account/webhooks).
    """

    def save(self):
        payload = {
            "name": self.name,
            "url": self.url,
            "events": self.events,
            "secret": self.secret,
            "active": self.active,
        }
        if self.id:
            url = host.endpoint(f"/workspaces/{self.workspace_id}/webhooks/{self.id}")
            response_json = self.session.put(url=url, json=payload).json()["data"]
            self.updated_at = response_json["updatedAt"]
            logger.info("Updated webhook %s", self)
        else:
            url = host.endpoint(f"/workspaces/{self.workspace_id}/webhooks")
            response_json = self.session.post(url=url, json=payload).json()["data"]
            self.id = response_json["id"]
            self.created_at = response_json["createdAt"]
            self.updated_at = response_json["updatedAt"]
            logger.info("Created webhook %s", self)

    @staticmethod
    def from_metadata(metadata: dict) -> "Webhook":
        return Webhook(
            id=metadata["id"],
            secret=metadata.get("secret"),
            active=metadata["active"],
            url=metadata["url"],
            name=metadata["name"],
            events=metadata["events"],
            created_at=metadata.get("createdAt"),
            updated_at=metadata.get("updatedAt"),
        )

    @classmethod
    def get(cls, webhook_id: str) -> "Webhook":
        url = host.endpoint(f"/workspaces/{cls.workspace_id}/webhooks/{webhook_id}")
        metadata = cls.session.get(url).json()["data"]
        return cls.from_metadata(metadata)

    def trigger_test_events(self) -> dict:
        """
        Triggers webhook test event to test your receiving side. The UP42 server will send test
        messages for each subscribed event to the specified webhook URL.

        Returns:
            A dict with information about the emitted test events.
        """
        url = host.endpoint(f"/workspaces/{self.workspace_id}/webhooks/{self.id}/tests")
        return self.session.post(url=url).json()["data"]

    def delete(self) -> None:
        """
        Deletes a registered webhook.
        """
        url = host.endpoint(f"/workspaces/{self.workspace_id}/webhooks/{self.id}")
        self.session.delete(url=url)
        logger.info("Successfully deleted Webhook: %s", self.id)

    @classmethod
    def get_webhook_events(cls) -> list[dict]:
        """
        Gets all available webhook events.

        Returns:
            A dict of the available webhook events.
        """
        url = host.endpoint("/webhooks/events")
        return cls.session.get(url=url).json()["data"]

    @classmethod
    def all(cls) -> List["Webhook"]:
        """
        Gets all registered webhooks for this workspace.

        Returns:
            A list of the registered webhooks for this workspace.
        """
        url = host.endpoint(f"/workspaces/{cls.workspace_id}/webhooks")
        payload = cls.session.get(url=url).json()["data"]
        logger.info("Queried %s webhooks.", len(payload))

        return [cls.from_metadata(metadata) for metadata in payload]



import dataclasses
from typing import Literal, Optional, Union

import geojson  # type: ignore

from up42 import base, host, order

UnitType = Literal["SQ_KM", "SCENE"]


@dataclasses.dataclass
class OrderError:
    index: int
    message: str
    details: str


@dataclasses.dataclass
class OrderReference:
    index: int
    id: str

    @property
    def order(self):
        return order.Order.get(self.id)


@dataclasses.dataclass
class OrderCost:
    index: int
    credits: float
    size: float
    unit: UnitType


@dataclasses.dataclass
class Estimate:
    items: list[Union[OrderCost, OrderError]]
    credits: float
    size: float
    unit: UnitType


def _get_items(data: dict, result_type):
    results = [result_type(**result) for result in data["results"]]
    errors = [OrderError(**error) for error in data["errors"]]
    items = results + errors
    return sorted(items, key=lambda x: x.index)


@dataclasses.dataclass
class BatchOrderTemplate:
    session = base.Session()
    workspace_id = base.WorkspaceId()
    data_product_id: str
    display_name: str
    features: geojson.FeatureCollection
    params: dict
    tags: Optional[list[str]] = None

    def __post_init__(self):
        self.__estimate()

    @property
    def _payload(self):
        payload = {
            "dataProduct": self.data_product_id,
            "displayName": self.display_name,
            "params": self.params,
            "featureCollection": self.features,
        }
        if self.tags is not None:
            payload["tags"] = self.tags
        return payload

    def __estimate(self):
        url = host.endpoint("/v2/orders/estimate")
        estimate = self.session.post(url=url, json=self._payload).json()
        summary = estimate["summary"]
        self.estimate = Estimate(
            items=_get_items(estimate, OrderCost),
            credits=summary["totalCredits"],
            size=summary["totalSize"],
            unit=summary["unit"],
        )

    def place(self) -> list[Union[OrderReference, OrderError]]:
        url = host.endpoint(f"/v2/orders?workspaceId={self.workspace_id}")
        batch = self.session.post(url=url, json=self._payload).json()
        return _get_items(batch, OrderReference)



import logging

from up42 import asset, catalog, order, storage, tasking, utils

logger = utils.get_logger(__name__, level=logging.INFO)

INITIALIZED_MSG = "Initialized %s"


@utils.deprecation(None, "3.0.0")
def initialize_catalog() -> catalog.Catalog:
    """
    Returns a Catalog object for using the catalog search.
    """
    return catalog.Catalog()


@utils.deprecation(None, "3.0.0")
def initialize_tasking() -> tasking.Tasking:
    """
    Returns a Tasking object for creating satellite tasking orders.
    """
    return tasking.Tasking()


@utils.deprecation(None, "3.0.0")
def initialize_storage() -> storage.Storage:
    """
    Returns a Storage object to list orders and assets.
    """
    return storage.Storage()


@utils.deprecation("Order::get", "3.0.0")
def initialize_order(order_id: str) -> order.Order:
    """
    Returns an Order object (has to exist on UP42).
    Args:
        order_id: The UP42 order_id
    """
    up42_order = order.Order.get(order_id)
    logger.info(INITIALIZED_MSG, up42_order)
    return up42_order


@utils.deprecation("Asset::get", "3.0.0")
def initialize_asset(asset_id: str) -> asset.Asset:
    """
    Returns an Asset object (has to exist on UP42).
    Args:
        asset_id: The UP42 asset_id
    """
    up42_asset = asset.Asset.get(asset_id)
    logger.info(INITIALIZED_MSG, up42_asset)
    return up42_asset



"""
Tasking functionality
"""

import dataclasses
import datetime
from typing import Any, Dict, Iterator, List, Literal, Optional, Union

from shapely import geometry as geom  # type: ignore

from up42 import base, catalog, glossary, host, order, utils

logger = utils.get_logger(__name__)

Geometry = Union[catalog.Geometry, geom.Point]

QuotationDecision = Literal[
    "ACCEPTED",
    "REJECTED",
]
QuotationStatus = Union[Literal["NOT_DECIDED"], QuotationDecision]
FeasibilityStatus = Literal["NOT_DECIDED", "ACCEPTED"]


class InvalidDecision(ValueError):
    pass


class Tasking(catalog.CatalogBase):
    """
    The Tasking class enables access to the UP42 tasking functionality.

    Use tasking:
    ```python
    tasking = up42.initialize_tasking()
    ```
    """

    def __init__(self):
        super().__init__(glossary.CollectionType.TASKING)

    @utils.deprecation("BatchOrderTemplate", "3.0.0")
    def construct_order_parameters(
        self,
        data_product_id: str,
        name: str,
        acquisition_start: Union[str, datetime.datetime],
        acquisition_end: Union[str, datetime.datetime],
        geometry: Geometry,
        tags: Optional[List[str]] = None,
    ) -> order.OrderParams:
        """
        Helps constructing the parameters dictionary required for the tasking order. Each sensor has additional
        parameters that are added to the output dictionary with value None. The potential values for to select from
        are given in the logs, for more detail on the parameter use `tasking.get_data_product_schema()`.

        Args:
            data_product_id: Id of the desired UP42 data product, see `tasking.get_data_products`
            name: Name of the tasking order project.
            acquisition_start: Start date of the acquisition period, datetime or isoformat string e.g. "2022-11-01"
            acquisition_end: End date of the acquisition period, datetime or isoformat string e.g. "2022-11-01"
            geometry: Geometry of the area to be captured, default a Polygon. Allows Point feature for specific
                data products. One of FeatureCollection, Feature, dict (geojson geometry), list (bounds coordinates),
                GeoDataFrame, shapely.Polygon, shapely.Point. All assume EPSG 4326!
            tags: A list of tags that categorize the order.

        Returns:
            The constructed order parameters dictionary.

        Example:
            ```python
            order_parameters = tasking.construct_order_parameters(
                data_product_id='647780db-5a06-4b61-b525-577a8b68bb54',
                name="My tasking order",
                acquisition_start=2022-11-01,
                acquisition_end=2022-12-01,
                geometry={'type': 'Polygon',
                   'coordinates': (((13.375966, 52.515068),
                     (13.375966, 52.516639),
                     (13.378314, 52.516639),
                     (13.378314, 52.515068),
                     (13.375966, 52.515068)),)}
                )
            ```
        """
        schema = self.get_data_product_schema(data_product_id)
        params: Dict[str, Any] = {param: None for param in schema["required"]}
        params |= {
            "displayName": name,
            "acquisitionStart": utils.format_time(acquisition_start),
            "acquisitionEnd": utils.format_time(acquisition_end, set_end_of_day=True),
        }
        order_parameters: order.OrderParams = {
            "dataProduct": data_product_id,
            "params": params,
        }
        if tags is not None:
            order_parameters["tags"] = tags
        geometry = utils.any_vector_to_fc(vector=geometry)
        if geometry["features"][0]["geometry"]["type"] == "Point":
            # Tasking (e.g. Blacksky) can require Point geometry.
            order_parameters["params"]["geometry"] = geometry["features"][0]["geometry"]
        else:
            geometry = utils.fc_to_query_geometry(fc=geometry, geometry_operation="intersects")
            order_parameters["params"]["geometry"] = geometry
        return order_parameters

    @utils.deprecation("Quotation::all", "3.0.0")
    def get_quotations(
        self,
        quotation_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        order_id: Optional[str] = None,
        decision: Optional[List[QuotationStatus]] = None,
        sortby: str = "createdAt",
        descending: bool = True,
    ) -> list[dict]:
        """
        This function returns the quotations for tasking by filtering and sorting by different parameters.

        Args:
            quotation_id (Optional[str], optional): The quotation Id for the specific quotation to retrieve.
            workspace_id (Optional[str], optional): The workspace id (uuid) to filter the search.
            order_id (Optional[str], optional): The order id (uuid) to filter the search.
            decision (Optional[list[str]], optional): quotation status (NOT_DECIDED, ACCEPTED or REJECTED).
            sortby: Param to sort the results. Default value set to createdAt
            descending (bool, optional): Descending or ascending sort.

        Returns:
            JSON: The json representation with the quotations resulted from the search.
        """
        params = {
            "workspaceId": workspace_id,
            "id": quotation_id,
            "orderId": order_id,
            "decision": decision,
            "sort": utils.SortingField(sortby, not descending),
        }
        return list(utils.paged_query(params, "/v2/tasking/quotation", self.session))

    @utils.deprecation("Quotation", "3.0.0")
    def decide_quotation(self, quotation_id: str, decision: QuotationDecision) -> dict:
        """Accept or reject a quotation for a tasking order.
        This operation is only allowed on quotations with the NOT_DECIDED status.

        Args:
            quotation_id (str): The target quotation ID.
            decision (str): The decision made for this quotation.

        Returns:
            dict: The confirmation to the decided quotation plus metadata.
        """
        if decision not in ["ACCEPTED", "REJECTED"]:
            raise InvalidDecision("Possible decisions are only ACCEPTED or REJECTED.")
        url = host.endpoint(f"/v2/tasking/quotation/{quotation_id}")
        return self.session.patch(url, json={"decision": decision}).json()

    @utils.deprecation(None, "3.0.0")
    def get_feasibility(
        self,
        feasibility_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        order_id: Optional[str] = None,
        decision: Optional[List[FeasibilityStatus]] = None,
        sortby: str = "createdAt",
        descending: bool = True,
    ) -> list[dict]:
        """
        This function returns the list of feasibility studies for tasking orders.

        Args:
            feasibility_id (Optional[str], optional): The feasibility Id for the specific feasibility study to retrieve.
            workspace_id (Optional[str], optional): The workspace id (uuid) to filter the search.
            order_id (Optional[str], optional): The order id (uuid) to filter the search.
            decision (Optional[list[str]], optional): The status of the quotation
            (NOT_DECIDED or ACCEPTED).
            sortby (str, optional): Arranges elements in asc or desc order based on a chosen field.
            The format is <field name>,<asc or desc>. Eg: "id,asc"
            descending (bool, optional): Descending or ascending sort.

        Returns:
            JSON: The json representation with the feasibility resulted from the search.
        """
        params = {
            "id": feasibility_id,
            "workspaceId": workspace_id,
            "orderId": order_id,
            "decision": decision,
            "sort": utils.SortingField(sortby, not descending),
        }
        return list(utils.paged_query(params, "/v2/tasking/feasibility-studies", self.session))

    @utils.deprecation(None, "3.0.0")
    def choose_feasibility(self, feasibility_id: str, accepted_option_id: str) -> dict:
        """Accept one of the proposed feasibility study options.
        This operation is only allowed on feasibility studies with the NOT_DECIDED status.

        Args:
            feasibility_id (str): The target feasibility study ID.
            accepted_option_id (str): The ID of the feasibility option to accept.

        Returns:
            dict: The confirmation to the decided quotation plus metadata.
        """
        url = host.endpoint(f"/v2/tasking/feasibility-studies/{feasibility_id}")
        return self.session.patch(url=url, json={"acceptedOptionId": accepted_option_id}).json()


class QuotationSorting:
    created_at = utils.SortingField(name="createdAt")
    credits_price = utils.SortingField(name="creditsPrice")
    decided_at = utils.SortingField(name="decisionAt")
    updated_at = utils.SortingField(name="updatedAt")


@dataclasses.dataclass
class Quotation:
    session = base.Session()
    id: str
    created_at: str
    updated_at: str
    decided_at: Optional[str]
    account_id: str
    workspace_id: str
    order_id: str
    credits_price: int
    decision: QuotationStatus

    def accept(self):
        self.decision = "ACCEPTED"

    def reject(self):
        self.decision = "REJECTED"

    def save(self):
        url = host.endpoint(f"/v2/tasking/quotation/{self.id}")
        metadata = self.session.patch(url, json={"decision": self.decision}).json()
        quotation = self._from_metadata(metadata)
        for field in dataclasses.fields(quotation):
            setattr(self, field.name, getattr(quotation, field.name))

    @staticmethod
    def _from_metadata(metadata: dict) -> "Quotation":
        return Quotation(
            id=metadata["id"],
            created_at=metadata["createdAt"],
            updated_at=metadata["updatedAt"],
            decided_at=metadata["decisionAt"],
            account_id=metadata["accountId"],
            workspace_id=metadata["workspaceId"],
            order_id=metadata["orderId"],
            credits_price=metadata["creditsPrice"],
            decision=metadata["decision"],
        )

    @classmethod
    def all(
        cls,
        quotation_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        order_id: Optional[str] = None,
        decision: Optional[List[QuotationStatus]] = None,
        sort_by: Optional[utils.SortingField] = None,
    ) -> Iterator["Quotation"]:
        params = {
            "workspaceId": workspace_id,
            "id": quotation_id,
            "orderId": order_id,
            "decision": decision,
            "sort": sort_by,
        }
        return map(
            cls._from_metadata,
            utils.paged_query(params, "/v2/tasking/quotation", cls.session),
        )
